 
from PyQt4 import QtCore, QtGui

class AuimdWidget(object):
  """The AuimdWidget provides applicative structure to widgets, which are user applications.

     All widgets share some basic things
     - two softkeys, which are QPushButton
        * they are displayed in the bottom bar
        * they are automatically connected to leftButtonAction/rightButtonAction
        * they can be freely manipulated by AuimdWidget subclasses
        
     - a title
        * usually displayed in the widget list
        * the default per class value can be redefined using defaultWidgetTitle
        * a specific value per widget instance (i.e. document name for a text
          editor) can be set using setWidgetTitle
          
     - an icon
        * same behaviour as title

     - a category
        * used to roughly classify widgets
        * for currently defined categories see core/widgets/mainmenu.py
        
     - a keyboard availability parameter
        * some widget may or may not want to display the on screen keyboard

  A widget must be constructed with the desktop as first parameter, and possibly
  arguments. These start arguments can be compared to command line parameters, are
  interpreted by the widget, and are not necessarily strings but can be any python
  objects.

  A widget can be specified as mono instance when not two instances can coexist
  at the same time (for instance a phone widget). If an attempt to start a mono
  instance widget is made multiple times, then the widget is put on top and the
  method restart is called (if defined).

  When creating a widget it must (at least) inherit from :
  - AuimdWidget : basic framework
  - a subclass of QWidget : UI
  """
  def __init__(self, parent, *args):
    self.desktop=parent.parent()
    
    self.args=args

    self.left_button=QtGui.QPushButton("")
    self.right_button=QtGui.QPushButton("")
    self.left_button.setFocusPolicy(QtCore.Qt.NoFocus)
    self.right_button.setFocusPolicy(QtCore.Qt.NoFocus)
    self.left_button.setFlat(1)
    self.right_button.setFlat(1)
    self.left_button.hide()
    self.right_button.hide()

    QtCore.QObject.connect(self.left_button, QtCore.SIGNAL("clicked()"), self.leftButtonAction)
    QtCore.QObject.connect(self.right_button, QtCore.SIGNAL("clicked()"), self.rightButtonAction)

    self.widget_icon=None
    self.widget_title=None

    self.virtual_size=None

  def leftButton(self):
    """Retrieve the left button (do not override, or with care)."""
    return self.left_button

  def rightButton(self):
    """Retrieve the right button (do not override, or with care)."""
    return self.right_button

  def leftButtonAction(self):
    """Called when the left button is clicked (override)."""
    pass

  def rightButtonAction(self):
    """Called when the right button is clicked (override)."""
    pass

  @staticmethod
  def defaultWidgetIcon():
    """Return the icon of the widget (override)."""
    return QtGui.QIcon("core/icons/gear.png")

  @staticmethod
  def defaultWidgetTitle():
    """Return the title of the widget (override)."""
    return "Untitled"

  @staticmethod
  def getCategory():
    """Return the category of the widget (override)."""
    return "Other"

  @staticmethod
  def monoInstance():
    """Specify whether only one instance of the widget can exist in the 
    same time.

    Default is False. If True a call to Desktop.startWidget will bring the 
    widget on top if one instance is already running.

    """
    return False

  def widgetIcon(self):
    """Return the icon of the widget (do not override)."""
    if(self.widget_icon):
      return self.widget_icon
    else:
      return self.defaultWidgetIcon()

  def widgetTitle(self):
    """Return the title of the widget (do not override)."""
    if(self.widget_title):
      return self.widget_title
    else:
      return self.defaultWidgetTitle()

  def setWidgetIcon(self, icon):
    """Set the icon of the widget (do not override)."""
    self.widget_icon=icon
    self.desktop.updateWidgetIconTitle(self)

  def setWidgetTitle(self, title):
    """Set the title of the widget (do not override)."""
    self.widget_title=title
    self.desktop.updateWidgetIconTitle(self)

  def keyboardAllowed(self):
    """The keyboard is allowed if this function returns 1 (override)."""
    return 1

  def readyToClose(self):
    """Called when the desktop would like to close the widget (override).

    Return 1 to accept closing, 0 to cancel the termination.

    """
    return 1

  def restart(self, *args):
    """Called when a mono instance widget is reactivated (override)."""
    pass
