"""Core Desktop class.

The Desktop is the central object in AUIMD. It manages components as below :

----------------------
|     Menu Bar       |
----------------------
|                    |
|                    |
|      Content       |
|     (widgets)      |
|                    |
|                    |
|                    |
----------------------
| On Screen keyboard |
|                    |
----------------------
|     Bottom Bar     |
----------------------

Widgets are AUIMD applications, they are usually shown in all the content space.
The desktop keeps track of the widgets in a list. The first widget is the
displayed widget.
The displayed widget is put in a scroll area, this enables to resize the widget
with an arbitrary size and be able to view all of its content with scrolling.
This simulate some sort of virtual space.

The desktop manages widgets life cycle :
- showWidget : show a widget, possibly adding it to the list of widget if not
               already in
- closeWidget : close a widget and remove it from the list of widgets. The
                next widget in the list is then shown
It is also possible to hide a widget with setWidgetVisibility.
The home widget is a little exception : it cannot be closed and the desktop has 
some particular knowledge of it. Appart from that, this is a regular widget.

Two kinds of widgets :
- those available in the core (home, mainmenu, exception)
- all other widgets
They share exactly the same structure but core widgets are mandatory, all other
are optional.
All widgets must inherit from AuimdWidgets AND a derived class of QWidget.
AuimdWidgets provides a basic applicative framework, whereas the QWidget derived
class provides graphical features. See core/auimd_widget.py for more.

The desktop also provides some utility function such as :
- open/save file dialog
- wallpaper management
- various plumbing

Basically most components will have a reference to the desktop, because it 
provides links between objects.

"""

import sys,os

from PyQt4 import QtCore, QtGui

import dbus
import dbus.service

from menu_bar import MenuBar
from bottom_bar import BottomBar
from keyboard import Keyboard
from core.window_catcher import WindowCatcher
from core.widgets.home import Home
from core.widgets.exception import Exception
from core.widgets.dialog_container import DialogContainer


def createDesktop(device, fullscreen, no_cursor, embed):
  """Create a class for creating a desktop instance.

  This indirection is used to choose between QtGui.QWidget QtGui.QX11EmbedWidget
  base class depending on the value of the embed argument.

  """

  if embed:
    base_class=QtGui.QX11EmbedWidget
  else:
    base_class=QtGui.QWidget

  class Desktop(base_class):
    NEW_XWINDOW_EVENT=QtCore.QEvent.User

    class DBusService(dbus.service.Object):
      """Handles requests made on the service exposed on the session bus."""
      def __init__(self, desktop):
        dbus.service.Object.__init__(self, dbus.SessionBus(), "/Desktop")
        self.desktop=desktop

      @dbus.service.method(dbus_interface='net.pierrox.auimd.Desktop', in_signature='sas', out_signature='i')
      def StartWidget(self, class_name, args):
        w=self.desktop.startWidget(class_name, *args)
        return w.winId()

      @dbus.service.method(dbus_interface='net.pierrox.auimd.Desktop', in_signature='i', out_signature='')
      def CloseWidget(self, winid):
        w=QtGui.QWidget.find(winid)
        if w:
          self.desktop.closeWidget(w)

      @dbus.service.method(dbus_interface='net.pierrox.auimd.Desktop', in_signature='i', out_signature='')
      def ShowWidget(self, winid):
        w=QtGui.QWidget.find(winid)
        if w:
          self.desktop.closeWidget(w)

    def __init__(self, device, fullscreen, no_cursor, embed):
      base_class.__init__(self, None)

      if fullscreen:
        self.setGeometry(QtGui.qApp.desktop().availableGeometry())
        self.setFixedSize(self.size())
      elif embed:
        self.embedInto(long(embed))

      # try to hide the cursor
      if no_cursor: 
        QtGui.qApp.setOverrideCursor(QtGui.QCursor(QtCore.Qt.BlankCursor))

      # load the device layer
      module=__import__("devices/"+device+"/device")
      self.device=module.Device()

      # check the screen size for possible desktop resize
      # (uses when no window manager present)
      if fullscreen:
        self.connect(QtGui.qApp.desktop(), QtCore.SIGNAL("resized(int)"), self.desktopResize)

      # create desktop components
      self.menu_bar=MenuBar(self)
      self.bottom_bar=BottomBar(self)
      self.content=QtGui.QWidget(self)
      self.content_scroll_h=QtGui.QScrollBar(QtCore.Qt.Horizontal, self)
      self.content_scroll_v=QtGui.QScrollBar(self)
      self.connect(self.content_scroll_h, QtCore.SIGNAL("valueChanged(int)"), self.scrollContentH)
      self.connect(self.content_scroll_v, QtCore.SIGNAL("valueChanged(int)"), self.scrollContentV)
      self.keyboard=Keyboard(self)
      self.keyboard.hide()

      # the list of widgets
      self.widgets=[]
      self.current_widget=None

      # add the home widget and show it
      self.home=Home(self.content)
      self.showWidget(self.home)

      # configuration data such as widgets directories, widget style
      self.loadSettings()

      # establish the list of discovered widgets
      self.loadWidgets()

      # start the "window catcher"
      self.window_catcher=WindowCatcher(self)
      self.window_catcher.start()

      # exception tracebacks are shown in a widget
      sys.excepthook=self.exceptionHandler

      # finally exposes DBus services
      self.dbus_iface=self.DBusService(self)

    def exceptionHandler(self, exc_type, exc_value, exc_traceback):
      """Display exceptions in a widget."""
      self.showWidget(Exception(self.content, exc_value, exc_traceback))

    def layoutComponents(self):
      """Layout components in a vertical box, the current widget being between the menu bar and the bottom bar.
      
      All this manual stuff is needed for generic embedding. Using a QScrollArea 
      breaks window reparenting. Also using a manual layout simplifies the 
      management of scrollbars (no asynchrone content resizing).
      """
      w=self.width()
      h=self.height()
      ms=self.menu_bar.minimumSizeHint()
      bs=self.bottom_bar.minimumSize()
      ks=self.keyboard.minimumSizeHint()
      rem_h=h

      self.menu_bar.setGeometry(0, 0, w, ms.height())
      rem_h-=ms.height()

      self.bottom_bar.setGeometry(0, h-bs.height(), w, bs.height())
      rem_h-=bs.height()

      if self.keyboard.isVisible():
        ksh=ks.height()
        self.keyboard.setGeometry(0, h-bs.height()-ksh, w, ksh)
        rem_h-=ksh

      if self.current_widget:
        if self.current_widget.virtual_size:
          ws=self.current_widget.virtual_size
        else:
          ws=self.current_widget.minimumSizeHint()

        if ws.width()>w or ws.height()>rem_h:
          sbw=16 # FIXME : find a correct way to cmpute scrollbar width
          if ws.width()>w:
            self.content_scroll_h.setGeometry(0, ms.height()+rem_h-sbw, w-sbw, sbw)
            self.content_scroll_h.setVisible(True)
            rem_h-=sbw
          else:
            self.content_scroll_h.setVisible(False)

          if ws.height()>rem_h:
            self.content_scroll_v.setGeometry(w-sbw, ms.height(), sbw, rem_h-sbw)
            self.content_scroll_v.setVisible(True)
            w-=sbw
          else:
            self.content_scroll_v.setVisible(False)

          self.content_scroll_h.setRange(0, ws.width()-w)
          self.content_scroll_v.setRange(0, ws.height()-rem_h)
          self.current_widget.move(0, 0)
          self.current_widget.resize(ws)
        else:
          self.content_scroll_h.setVisible(False)
          self.content_scroll_v.setVisible(False)
          self.current_widget.setGeometry(0, 0, w, rem_h)
        self.content.setGeometry(0, ms.height(), w, rem_h)

    def scrollContentH(self, value):
      self.current_widget.move(-value, self.current_widget.y())

    def scrollContentV(self, value):
      self.current_widget.move(self.current_widget.x(), -value)

    def desktopResize(self, screen):
      """Resize the desktop according to the screen size.

      This function is only called when in fullscreen mode.

      """
      r=QtGui.qApp.desktop().availableGeometry()
      self.setGeometry(r)
      self.setFixedSize(r.size())
      
    def resizeEvent(self, e=None):
      """Resize event are generated when in windowed mode."""
      self.layoutComponents()

    def showHideKeyboard(self, state=None):
      """Show or hide the keyboard, depending on its current state.

      If state is None, invert the state. Otherwise use the state value.

      """
      if(state==None):
        if(self.keyboard.isVisible()):
          state=0
        else:
          state=1
    
      if(state):
        self.keyboard.show()
      else:
        self.keyboard.hide()
      self.layoutComponents()
      
    def startWidget(self, class_name, *args):
      """Start then show a widget by its class name, with optional arguments.
      
      If a class is single instance, and an instance is already running, it
      is simply put on top.
      
      """

      #if class_name not in self.widget_class_names:
      #  return False

      # get a class according to its name
      wc=self.widget_class_names[class_name]

      # check whether this is a mono or multi instances class, and if mono check for
      # an already running instance
      if wc.monoInstance():
        l=[x for x in self.widgets if x.__class__.__name__==class_name]
        if len(l)>0:
          w=l[0]
          w.restart(*args)
        else:
          w=wc(self.content, *args)
      else:
        w=wc(self.content, *args)

      # finally show and return the widget
      self.showWidget(w)
      return w


    def showWidget(self, widget):
      """Show a widget, and add it to the widgets list if not already in.
      
      In all cases the widget is put in front of the list. It is also resized
      according to its virtual size if needed. Showing a widget also shows its
      buttons.
      
      """
      if(widget in self.widgets):
        self.widgets.remove(widget)
      self.widgets.insert(0, widget)

      # replace the actually visible widget (if any) by the new one
      if self.current_widget:
        self.current_widget.hide()

      self.current_widget=widget

      self.layoutComponents()

      widget.show()

      widget.activateWindow()
      
      # enable the keyboard only if the widget accepts it
      self.bottom_bar.setButtons(widget.leftButton(), widget.rightButton(), widget.keyboardAllowed())
      if(not widget.keyboardAllowed()): self.showHideKeyboard(0)

      # update the menubar
      self.menu_bar.setCloseButtonEnabled(not self.current_widget is self.home)
      self.menu_bar.selectWidget(widget)

    def closeCurrentWidget(self):
      """Close the current widget (the displayed widget) and show the next."""
      if len(self.widgets)==1: return
      self.closeWidget(self.current_widget)

    def closeWidget(self, old):
      """Close a given widget, and show the next."""
      if old not in self.widgets: return

      if not old.readyToClose(): return

      self.widgets.remove(old)
      self.menu_bar.closeWidget(old)

      self.showWidget(self.widgets[0])
      
      old.close()
      old.deleteLater()

    def setWidgetVisibility(self, widget, visibility):
      """Show or hide a widget without closing it."""
      self.menu_bar.setWidgetVisibility(widget, visibility)

    def updateWidgetIconTitle(self, widget):
      """Propagate widget's icon and title changes."""
      self.menu_bar.updateWidgetIconTitle(widget)

    def dialog(self, opener, dialog, title=None, left_button_text="Ok", right_button_text="Cancel", icon=None):
      """Show a dialog, the current widget is hidden and replaced by the dialog, 
      the caller is blocked until the dialog is finished or closed.
      
      The dialog argument must be a subclass of QtGui.QDialog.
      Other widgets are not blocked.
      The function return Accepted or Rejected, or None if the dialog is closed.

      """
      if not icon:
        icon=opener.widgetIcon()

      if not title:
        title=opener.widgetTitle()
      else:
        title=opener.widgetTitle()+" - "+title

      w=DialogContainer(self.content, dialog, title, left_button_text, right_button_text, icon)
      self.showWidget(w)
      self.setWidgetVisibility(opener, 0)
      self.layoutComponents()
      result=w.exec_()
      self.setWidgetVisibility(opener, 1)
      self.closeWidget(w)
      self.layoutComponents()
      return result

    def openFile(self, opener):
      """Shown a file picker for opening.
      
      opener is the widget at the origin of the request. It will be hidden
      while the file is being picked.
      Return the filename or None.
      
      """
      d=QtGui.QFileDialog()
      d.setAcceptMode(QtGui.QFileDialog.AcceptOpen)
      result=self.dialog(opener, d, "Open...", "Open", "Cancel", QtGui.QIcon("core/icons/fileopen.png"))
      if result==QtGui.QDialog.Accepted:
        return str(d.selectedFiles()[0])
      else:
        return None

    def saveFile(self, opener):
      """Shown a file picker for saving.
      
      opener is the widget at the origin of the request. It will be hidden
      while the file is being picked.
      Return the filename or None.
      
      """
      d=QtGui.QFileDialog()
      d.setAcceptMode(QtGui.QFileDialog.AcceptSave)
      result=self.dialog(opener, d, "Save...", "Save", "Cancel", QtGui.QIcon("core/icons/filesave.png"))
      if result==QtGui.QDialog.Accepted:
        return str(d.selectedFiles()[0])
      else:
        return None

    def chooseDirectory(self, opener):
      """Shown a file picker for directory choosing.
      
      opener is the widget at the origin of the request. It will be hidden
      while the directory is being choosen.
      Return the directory name or None.
      
      """
      d=QtGui.QFileDialog()
      d.setFileMode(QtGui.QFileDialog.DirectoryOnly)
      d.setAcceptMode(QtGui.QFileDialog.AcceptOpen)
      result=self.dialog(opener, d, "Choose a directory...", "Choose", "Cancel", QtGui.QIcon("core/icons/fileopen.png"))
      if result==QtGui.QDialog.Accepted:
        return str(d.selectedFiles()[0])
      else:
        return None

    def setWallpaper(self, path, scaled):
      """Set the home wallpaper attributes."""
      self.home.setWallpaper(path, scaled)

    def getWallpaper(self):
      """Retrieve the wallpaper of the home widget."""
      return self.home.wallpaper

    def closeEvent(self, e):
      """Try to shut down things cleanly."""
      self.keyboard.close()
      for i in self.widgets: 
        i.close()

    def customEvent(self, e):
      if e.type()==Desktop.NEW_XWINDOW_EVENT:
        self.startWidget('XWindow', e.id, e.name)

    def exit(self):
      """Request the application to terminate."""
      self.close()

    def loadWidgets(self, do_reload=False):
      """Look recursively in widgets directories folder for valid widgets classes.
      
      Search for packages, where __init__.py contains a AuimdWidgets tuple,
      giving the list of widgets provided by the package.
      
      """

      # hold classes of widgets that have been discovered (dict category -> list of classes)
      self.widget_classes={}

      # index of all widgets by name (dict name -> class)
      self.widget_class_names={}

      categories=("Graphics", "Games", "Multimedia", "Tools", "Network", "PIM", "Settings", "System", "Other", "Hidden")
      for c in categories:
        self.widget_classes[c]=[]

      for directory in self.widgets_directories:
        for dirpath, dirs, files in os.walk(directory):
          if "__init__.py" in files:
            if dirpath[0]=="/":
              sys.path.insert(0, os.path.dirname(dirpath))
              m=__import__(os.path.basename(dirpath))
            else:
              m=__import__(dirpath)
            try:
              if do_reload:
                reload(m)
                # this way to manually reload widgets is not completely satisfying because we need to list exhaustively all required modules for a widget
                modules=getattr(m, "AuimdModules", None)
                if modules:
                  for i in modules:
                    reload(i)
                # a second reload is needed because AuimdWidgets references modules before their reload
                reload(m)
              auimd_widgets=getattr(m, "AuimdWidgets")
              for w in auimd_widgets:
                c=w.getCategory()
                if not c in categories and c != "Hidden":
                  c="Other"
                self.widget_classes[c].append(w)
                self.widget_class_names[w.__name__]=w
                self.widget_packages.append(m)
            except AttributeError:
              # does not look like an auimd package
              pass

            # remove a search path that could have been inserted in cas of absolute path
            if dirpath[0]=="/":
              del sys.path[0]

    def reloadWidgets(self):
      """Reload widgets (after a first use of loadWidgets)."""
      self.loadWidgets(True)

    def loadSettings(self):
      settings=QtCore.QSettings("auimd", "Desktop")
      self.loadStyleSetting(settings)
      self.loadWidgetsDirectories(settings)
    
    def loadStyleSetting(self, settings):
      style=settings.value("style", QtCore.QVariant("cleanlooks")).toString()
      QtGui.qApp.setStyle(style)

    def saveStyleSetting(self):
      settings=QtCore.QSettings("auimd", "Desktop")
      settings.setValue("style", QtCore.QVariant(QtGui.qApp.style().objectName()))
      
    def loadWidgetsDirectories(self, settings):
      # always put the "built-in" widgets directories in front
      self.widgets_directories=["widgets"]
      n=settings.beginReadArray("widgets_directories")
      for i in range(n):
        settings.setArrayIndex(i)
        self.widgets_directories.append(str(settings.value("directory").toString()))
      settings.endArray()

    def saveWidgetsDirectories(self):
      settings=QtCore.QSettings("auimd", "Desktop")
      settings.beginWriteArray("widgets_directories")
      # the first element ("widgets") is skipped (built-in)
      for i in range(len(self.widgets_directories)-1):
        settings.setArrayIndex(i)
        settings.setValue("directory", QtCore.QVariant(self.widgets_directories[i+1]))
      settings.endArray()

  return Desktop(device, fullscreen, no_cursor, embed)
